@@ -4,25 +4,21 @@
  *
  * Thanks to Niki W. Waibel <niki.waibel@gmx.net> for a prototype version
  *
-    This library is free software; you can redistribute it and/or
-    modify it under the terms of the GNU Lesser General Public
-    License as published by the Free Software Foundation; either
-    version 2.1 of the License, or (at your option) any later version.
+	This library is free software; you can redistribute it and/or
+	modify it under the terms of the GNU Lesser General Public
+	License as published by the Free Software Foundation; either
+	version 2.1 of the License, or (at your option) any later version.
 
-    This library is distributed in the hope that it will be useful,
-    but WITHOUT ANY WARRANTY; without even the implied warranty of
-    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
-    Lesser General Public License for more details.
+	This library is distributed in the hope that it will be useful,
+	but WITHOUT ANY WARRANTY; without even the implied warranty of
+	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+	Lesser General Public License for more details.
 
 	You should have received a copy of the GNU Lesser General Public License
 	along with this library; if not, write to the Free Software Foundation,
 	Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
  */
 
-/*
- * $Id: ccid_serial.c 6021 2011-10-10 13:00:23Z rousseau $
- */
-
 #include <stdio.h>
 #include <stdlib.h>
 #include <fcntl.h>
@@ -35,7 +31,7 @@
 #include <sys/ioctl.h>
 #include <ifdhandler.h>
 
-#include "config.h"
+#include <config.h>
 #include "defs.h"
 #include "ccid_ifdhandler.h"
 #include "debug.h"
@@ -154,6 +150,9 @@ unsigned int SerialExtendedDataRates[] = { ISO_DATA_RATES, 500000, 0 };
 /* data rates supported by the secondary slots on the GemCore Pos Pro & SIM Pro */
 unsigned int SerialCustomDataRates[] = { GEMPLUS_CUSTOM_DATA_RATES, 0 };
 
+/* data rates supported by the GemCore SIM Pro 2 */
+unsigned int SIMPro2DataRates[] = { SIMPRO2_ISO_DATA_RATES, 0  };
+
 /* no need to initialize to 0 since it is static */
 static _serialDevice serialDevice[CCID_DRIVER_MAX_READERS];
 
@@ -386,7 +385,11 @@ int get_bytes(unsigned int reader_index, unsigned char *buffer, int length)
 		rv = ReadChunk(reader_index, serialDevice[reader_index].buffer,
 			sizeof(serialDevice[reader_index].buffer), length - present);
 		if (rv < 0)
+		{
+			serialDevice[reader_index].buffer_offset = 0;
+			serialDevice[reader_index].buffer_offset_last = 0;
 			return STATUS_COMM_ERROR;
+		}
 
 		/* fill the buffer */
 		memcpy(buffer + present, serialDevice[reader_index].buffer,
@@ -522,8 +525,12 @@ static status_t set_ccid_descriptor(unsigned int reader_index,
 		readerID = GEMCOREPOSPRO;
 	else if (0 == strcasecmp(reader_name,"GemCoreSIMPro"))
 		readerID = GEMCORESIMPRO;
+	else if (0 == strcasecmp(reader_name,"GemCoreSIMPro2"))
+		readerID = GEMCORESIMPRO2;
 	else if (0 == strcasecmp(reader_name,"GemPCPinPad"))
 		readerID = GEMPCPINPAD;
+	else if (0 == strcasecmp(reader_name,"SEC1210"))
+		readerID = SEC1210;
 
 	/* check if the same channel is not already used to manage multi-slots readers*/
 	for (i = 0; i < CCID_DRIVER_MAX_READERS; i++)
@@ -558,10 +565,26 @@ static status_t set_ccid_descriptor(unsigned int reader_index,
 			{
 				case GEMCOREPOSPRO:
 				case GEMCORESIMPRO:
-					serialDevice[reader_index].ccid.arrayOfSupportedDataRates = SerialCustomDataRates;
+					{
+						/* Allocate a memory buffer that will be
+						 * released in CloseUSB() */
+						void *ptr = malloc(sizeof SerialCustomDataRates);
+						if (ptr)
+						{
+							memcpy(ptr, SerialCustomDataRates,
+									sizeof SerialCustomDataRates);
+						}
+
+						serialDevice[reader_index].ccid.arrayOfSupportedDataRates = ptr;
+					}
 					serialDevice[reader_index].ccid.dwMaxDataRate = 125000;
 					break;
 
+				case SEC1210:
+					serialDevice[reader_index].ccid.arrayOfSupportedDataRates = NULL;
+					serialDevice[reader_index].ccid.dwMaxDataRate = 826000;
+					break;
+
 				/* GemPC Twin or GemPC Card */
 				default:
 					serialDevice[reader_index].ccid.arrayOfSupportedDataRates = SerialTwinDataRates;
@@ -598,8 +621,14 @@ static status_t set_ccid_descriptor(unsigned int reader_index,
 	serialDevice[reader_index].ccid.dwMaxDataRate = 344086;
 	serialDevice[reader_index].ccid.bMaxSlotIndex = 0;
 	serialDevice[reader_index].ccid.arrayOfSupportedDataRates = SerialTwinDataRates;
+	serialDevice[reader_index].ccid.readTimeout = DEFAULT_COM_READ_TIMEOUT;
 	serialDevice[reader_index].ccid.dwSlotStatus = IFD_ICC_PRESENT;
 	serialDevice[reader_index].ccid.bVoltageSupport = 0x07;	/* 1.8V, 3V and 5V */
+	serialDevice[reader_index].ccid.gemalto_firmware_features = NULL;
+	serialDevice[reader_index].ccid.dwProtocols = SCARD_PROTOCOL_T0 | SCARD_PROTOCOL_T1;
+#ifdef ENABLE_ZLP
+	serialDevice[reader_index].ccid.zlp = FALSE;
+#endif
 	serialDevice[reader_index].echo = TRUE;
 
 	/* change some values depending on the reader */
@@ -619,11 +648,29 @@ static status_t set_ccid_descriptor(unsigned int reader_index,
 			serialDevice[reader_index].ccid.dwMaxDataRate = 500000;
 			break;
 
+		case GEMCORESIMPRO2:
+			serialDevice[reader_index].ccid.dwDefaultClock = 4800;
+			serialDevice[reader_index].ccid.bMaxSlotIndex = 1; /* 2 slots */
+			serialDevice[reader_index].ccid.arrayOfSupportedDataRates = SIMPro2DataRates;
+			serialDevice[reader_index].echo = FALSE;
+			serialDevice[reader_index].ccid.dwMaxDataRate = 825806;
+			break;
+
 		case GEMPCPINPAD:
 			serialDevice[reader_index].ccid.bPINSupport = 0x03;
 			serialDevice[reader_index].ccid.arrayOfSupportedDataRates = SerialExtendedDataRates;
 			serialDevice[reader_index].ccid.dwMaxDataRate = 500000;
 			break;
+
+		case SEC1210:
+			serialDevice[reader_index].ccid.dwFeatures = 0x000100B2;
+			serialDevice[reader_index].ccid.dwDefaultClock = 4800;
+			serialDevice[reader_index].ccid.dwMaxDataRate = 826000;
+			serialDevice[reader_index].ccid.arrayOfSupportedDataRates = NULL;
+			serialDevice[reader_index].ccid.bMaxSlotIndex = 1;	/* 2 slots */
+			serialDevice[reader_index].echo = FALSE;
+			break;
+
 	}
 
 end:
@@ -737,10 +784,62 @@ status_t OpenSerialByName(unsigned int reader_index, char *dev_name)
 	 * will echo characters for you.  Don't generate signals. */
 	current_termios.c_lflag = 0;
 
+	if (0 == strcasecmp(reader_name,"GemCoreSIMPro2"))
+	{
+		unsigned char pcbuffer[SIZE_GET_SLOT_STATUS];
+		unsigned int old_timeout;
+		RESPONSECODE r;
+
+		/* Unless we resume from a stand-by condition, GemCoreSIMPro2
+		 * starts at 9600 bauds, so let's first try this speed */
+		/* set serial port speed to 9600 bauds */
+		(void)cfsetspeed(&current_termios, B9600);
+		DEBUG_INFO1("Set serial port baudrate to 9600 and correct configuration");
+		if (tcsetattr(serialDevice[reader_index].fd, TCSANOW, &current_termios) == -1)
+		{
+			(void)close(serialDevice[reader_index].fd);
+			serialDevice[reader_index].fd = -1;
+			DEBUG_CRITICAL2("tcsetattr error: %s", strerror(errno));
+
+			return STATUS_UNSUCCESSFUL;
+		}
+
+		/* Test current speed issuing a CmdGetSlotStatus with a very
+		 * short time out of 1 seconds */
+		old_timeout = serialDevice[reader_index].ccid.readTimeout;
+
+		serialDevice[reader_index].ccid.readTimeout = 1*1000;
+		r = CmdGetSlotStatus(reader_index, pcbuffer);
+
+		/* Restore default time out value */
+		serialDevice[reader_index].ccid.readTimeout = old_timeout;
+
+		if (IFD_SUCCESS == r)
+		{
+			/* We are at 9600 bauds, let's move to 115200 */
+			unsigned char tx_buffer[] = { 0x01, 0x10, 0x20 };
+			unsigned char rx_buffer[50];
+			unsigned int rx_length = sizeof(rx_buffer);
+
+			if (IFD_SUCCESS == CmdEscape(reader_index, tx_buffer,
+				sizeof(tx_buffer), rx_buffer, &rx_length, 0))
+			{
+				/* Let the reader setup its new communication speed */
+				(void)usleep(250*1000);
+			}
+			else
+			{
+				DEBUG_INFO1("CmdEscape to configure 115200 bauds failed");
+			}
+		}
+		/* In case of a failure, reader is probably already at 115200
+		 * bauds as code below assumes */
+	}
+
 	/* set serial port speed to 115200 bauds */
 	(void)cfsetspeed(&current_termios, B115200);
 
-	DEBUG_INFO("Set serial port baudrate to 115200 and correct configuration");
+	DEBUG_INFO1("Set serial port baudrate to 115200 and correct configuration");
 	if (tcsetattr(serialDevice[reader].fd, TCSANOW, &current_termios) == -1)
 	{
 		(void)close(serialDevice[reader].fd);
@@ -753,24 +852,24 @@ status_t OpenSerialByName(unsigned int reader_index, char *dev_name)
 	/* perform a command to be sure a Gemalto reader is connected
 	 * get the reader firmware */
 	{
-		unsigned char tx_buffer[] = { 0x02 };
+		unsigned char tx_buffer[1];
 		unsigned char rx_buffer[50];
 		unsigned int rx_length = sizeof(rx_buffer);
 
-		/* 2 seconds timeout to not wait too long if no reader is connected */
-		serialDevice[reader].ccid.readTimeout = 2*1000;
+		if (0 == strcasecmp(reader_name,"SEC1210"))
+			tx_buffer[0] = 0x06; // unknown but supported command
+		else
+			tx_buffer[0] = 0x02; // get reader firmware
 
+		/* 2 seconds timeout to not wait too long if no reader is connected */
 		if (IFD_SUCCESS != CmdEscape(reader_index, tx_buffer, sizeof(tx_buffer),
-			rx_buffer, &rx_length))
+			rx_buffer, &rx_length, 2*1000))
 		{
 			DEBUG_CRITICAL("Get firmware failed. Maybe the reader is not connected");
 			(void)CloseSerial(reader_index);
 			return STATUS_UNSUCCESSFUL;
 		}
 
-		/* normal timeout: 2 seconds */
-		serialDevice[reader].ccid.readTimeout = DEFAULT_COM_READ_TIMEOUT ;
-
 		rx_buffer[rx_length] = '\0';
 		DEBUG_INFO2("Firmware: %s", rx_buffer);
 	}
@@ -778,13 +877,14 @@ status_t OpenSerialByName(unsigned int reader_index, char *dev_name)
 	/* perform a command to configure GemPC Twin reader card movement
 	 * notification to synchronous mode: the card movement is notified _after_
 	 * the host command and _before_ the reader anwser */
+	if (0 != strcasecmp(reader_name,"SEC1210"))
 	{
 		unsigned char tx_buffer[] = { 0x01, 0x01, 0x01};
 		unsigned char rx_buffer[50];
 		unsigned int rx_length = sizeof(rx_buffer);
 
 		if (IFD_SUCCESS != CmdEscape(reader_index, tx_buffer, sizeof(tx_buffer),
-			rx_buffer, &rx_length))
+			rx_buffer, &rx_length, 0))
 		{
 			DEBUG_CRITICAL("Change card movement notification failed.");
 			(void)CloseSerial(reader_index);
